// pingpong2_noref.cpp: two ping-pong parties (without reference types)
// RD, 04/24/18

#include <stdio.h>
#include <assert.h>
#include "systemc.h"

#define ITERATIONS (42*10)
#define WORKLOAD   1000000

// deterministic MT-safe pseudo random number in range [0...Max]
// (based on example by 'man rand', 4/8/17, RD)
int Rnd(unsigned long *Next, int Max)
{
    *Next = *Next * 1103515245 + 12345;
    return((unsigned)(*Next/65536) % (Max+1));
}

// simulated workload

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

// user-defined data type

typedef struct Data_t
{
  double 	x, y;

  inline friend std::ostream& operator<<(
	std::ostream& os, const Data_t& d)
  { os << d.x << ' ' << d.y << '\n';
    return os;
  }
} data_t;

// user-defined double-handshake channel (with transceiver interface)

class i_data_t: virtual public sc_interface
{
    public:
	virtual data_t read(void) = 0;
	virtual void write(const data_t) = 0;
};

class c_double_handshake:
	public sc_channel, public i_data_t
{
    public:
	c_double_handshake(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_double_handshake()
	{ }

	data_t read(void)
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    data_t temp = tmpd;
	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	void write(const data_t X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}

    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	data_t tmpd;
};

// Ping

SC_MODULE(Ping)
{
    sc_port<i_data_t> Port;
    unsigned Count;
    unsigned long w;

    void main(void);

    SC_CTOR(Ping)
    {
	Count = 0;
	SC_HAS_PROCESS(Ping);
	SC_THREAD(main);
    }
};

void Ping::main(void)
{
    data_t p;
    int i;
    unsigned long RndNext = 1;

    for(i=0; i<ITERATIONS; i++)
    {
	p.x = i;
	p.y = i;
	Port->write(p);
	Count++;
	w = Work(Rnd(&RndNext, WORKLOAD));
//	wait(1, SC_NS);
	p = Port->read();
	assert(p.x == -i);
	assert(p.y == -i);
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Pong

SC_MODULE(Pong)
{
    sc_port<i_data_t> Port;
    unsigned Count;
    unsigned long w;

    void main(void);

    SC_CTOR(Pong)
    {
	Count = 0;
	SC_HAS_PROCESS(Pong);
	SC_THREAD(main);
    }
};

void Pong::main(void)
{
    data_t p;
    int i;
    unsigned long RndNext = 2;

    for(i=0; i<ITERATIONS; i++)
    {
	p = Port->read();
	assert(p.x == i);
	assert(p.y == i);
	p.x *= -1.0;
	p.y *= -1.0;
	w = Work(Rnd(&RndNext, WORKLOAD));
//	wait(1, SC_NS);
	Port->write(p);
	Count++;
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Top

SC_MODULE(TOP)
{
    c_double_handshake c1, c2;
    Ping ping1, ping2;
    Pong pong1, pong2;

    SC_CTOR(TOP)
    :c1("c1"), ping1("ping1"), pong1("pong1"),
     c2("c2"), ping2("ping2"), pong2("pong2")
    {
    }

    void before_end_of_elaboration()
    {
	ping1.Port.bind(c1);
	pong1.Port.bind(c1);
	ping2.Port.bind(c2);
	pong2.Port.bind(c2);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("Ping-pong starts\n");
    sc_start();
    assert(top.ping1.Count == ITERATIONS);
    assert(top.pong1.Count == ITERATIONS);
    assert(top.ping2.Count == ITERATIONS);
    assert(top.pong2.Count == ITERATIONS);
    printf("Ping-pong done.\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
