#include "rose.h"
#include "risc.h"

#include "expression_analysis.h"

#include <iostream>
#include <fstream>

#include <list>
#include <set>
#include <algorithm>

extern std::set<SgFunctionDefinition*> future_functions;
std::list<std::string> output;

#define HELP_PAGE                                                                   \
"simd: " RISC_LIBRARY_NAME " SIMD Advisor, V" RISC_VERSION_NUMBER "\n"              \
"      " RISC_COPYRIGHT "\n"                                                        \
"\n"                                                                                \
"Usage: simd [<option>...] <design> [<option>...]\n"                                \
"\n"                                                                                \
"Options:               Description:                    Default:\n"                 \
"-h|--help              print this usage info, quit     none\n"                     \
"-v|--verbose           increment verbosity             silent\n"                   \
"-vv                    increment verbosity by 2        silent\n"                   \
"-vvv                   increment verbosity by 3        silent\n"                   \
"-w|--warnings          increment warning level         no warnings\n"              \
"-ww                    increment warning level by 2    no warnings\n"              \
"-www                   increment warning level by 3    no warnings\n"              \
"-I<dir>                add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-o <filename>          name output file as filename    none\n"                     \
"<rose:option>          passed to ROSE compiler         none\n"                     \
"<g++_option>           passed to GNU C++ compiler      none\n"                     \
"\n"

class RISCParseStatus: public risc::tools::ParseStatus
{
  public:
    bool dump;
    std::string filename;

    RISCParseStatus(): ParseStatus(), dump(false) {}
    RISCParseStatus& operator=(const RISCParseStatus& copy)
    {
      verbosityLevel = copy.verbosityLevel;
      warningLevel = copy.warningLevel;
      compilerName = copy.compilerName;

      return *this;
    }
};

void
RISCParseCommandlineOptions(int &argc, char** &argv, RISCParseStatus* status)
{
  if (argc < 2)
  {
    std::cout << "simd: no input files\n";
    exit(0);
  }

  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    // -h, --help
    if (it->compare("-h") == 0 || it->compare("--help") == 0)
    {
      std::cout << HELP_PAGE;
      exit(0);
    }

    if (it->compare("-o") == 0)
    {
      it++;
      if(it == argList.end()) {
        std::cout << "Flag -o has no filename" << std::endl;
        exit(1);
      }
      status->filename = *it;
      status->dump = true;
    }
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);

  // Parse library-level command-line options
  risc::tools::ParseCommandlineOptions(argc, argv, status);
}

enum SimdCompatibility {Supported=0, WrongArrayIndex=1, VariableConflict=2, Unsupported=3};
// Unsupported: writing not local var, continue, goto statement

void find_simd_loops(risc::Design &d);
SimdCompatibility is_simd_compatible(SgStatement *stmt, bool is_in_loop, std::list<SgNode*> &expressions);

SgVarRefExp* writing_belongs_to_array(SgNode *node, SgVariableSymbol *var_symbol)
{
  if(isSgVariableDeclaration(node)) {
    return NULL; // WAS: return false;  (RD, 05/25/18)
  }

  SgExprStatement *stmt_expr = isSgExprStatement(node);

  SgExpression *expr = NULL;
  if(stmt_expr) {
    expr = stmt_expr->get_expression();
  }

  if(expr == NULL) {
    expr = isSgExpression(node);
  }

  std::vector<SgNode*> array_access = NodeQuery::querySubTree(expr, V_SgVarRefExp);

  for(std::vector<SgNode*>::iterator
      iter  = array_access.begin();
      iter != array_access.end();
      iter++) {

    SgDotExp *dot = isSgDotExp((*iter)->get_parent());
    SgArrowExp *arrow = isSgArrowExp((*iter)->get_parent());

    if(dot) {

      SgPntrArrRefExp *pntr_array = isSgPntrArrRefExp(dot->get_lhs_operand());

      if(pntr_array) {

        SgArrowExp *arrow2 = isSgArrowExp(pntr_array->get_lhs_operand());
        if(arrow2) {
          if(isSgVarRefExp(arrow2->get_rhs_operand())) {
            return isSgVarRefExp(arrow2->get_rhs_operand());
          }
        }
        return isSgVarRefExp(pntr_array->get_lhs_operand());
      }
    }

    if(arrow) {

      if(isSgPntrArrRefExp(arrow->get_parent())) {
        return isSgVarRefExp(*iter);
      }
    }

    if(isSgPntrArrRefExp((*iter)->get_parent())) {
      return isSgVarRefExp(isSgPntrArrRefExp((*iter)->get_parent())->get_lhs_operand());
    }
  }

  return NULL;
}

bool is_declared_in_loop(SgVariableSymbol *var_symbol, SgScopeStatement *scope)
{
  std::vector<SgNode*> local_variables
    = NodeQuery::querySubTree(scope, V_SgVariableDeclaration);

  for(std::vector<SgNode*>::iterator
      iter  = local_variables.begin();
      iter != local_variables.end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

    if(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table() == var_symbol) {
      return true;
    }
  }
  return false;
}

SimdCompatibility writes_only_simd_compatible_variables_while_loop(std::set<SgVariableSymbol*> &write, SgWhileStmt *while_stmt)
{
  for(std::set<SgVariableSymbol*>::iterator
      iter  = write.begin();
      iter != write.end();
      iter++ ) {

    if(risc::tools::is_member_variable(*iter) && !isSgArrayType((*iter)->get_type())) {
      return VariableConflict;
    }

    if(risc::tools::is_local_variable(*iter)) {

      if(!isSgScopeStatement(while_stmt->get_body())->symbol_exists(*iter)) {

        return VariableConflict;
      }
    }
  }

  return Supported;
}

SimdCompatibility writes_only_simd_compatible_variables(std::list<SgNode*> &expressions, SgVariableSymbol *increment_var, SgScopeStatement *for_stmt)
{
  SimdCompatibility result = Supported;

  for(std::list<SgNode*>::iterator
      expr_iter  = expressions.begin();
      expr_iter != expressions.end();
      expr_iter++) {

    std::set<SgVariableSymbol*> read_variables;
    std::set<SgVariableSymbol*> write_variables;

    SgStatement *stmt = isSgStatement(*expr_iter);
    analyze_statement(stmt, read_variables, write_variables);

    for(std::set<SgVariableSymbol*>::iterator
        write_symbol_iter  = write_variables.begin();
        write_symbol_iter != write_variables.end();
        write_symbol_iter++) {

      bool is_local_variable_in_loop = false;
      std::vector<SgNode*> local_variables
        = NodeQuery::querySubTree(*expr_iter, V_SgVariableDeclaration);

      for(std::vector<SgNode*>::iterator
          var_decl_iter  = local_variables.begin();
          var_decl_iter != local_variables.end();
          var_decl_iter++) {

        SgVariableDeclaration *var_decl = isSgVariableDeclaration(*var_decl_iter);
        if(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table() == (*write_symbol_iter)) {
          is_local_variable_in_loop = true;
        }
      }

      SgVarRefExp *array_var = writing_belongs_to_array(stmt, *write_symbol_iter);

      if(array_var) {

        if(SageInterface::getDimensionCount(array_var->get_type()) != 1) {

          result = Unsupported;
          continue;
        }

        std::vector<SgNode*> array_access
          = NodeQuery::querySubTree(*expr_iter, V_SgPntrArrRefExp);

        for(std::vector<SgNode*>::iterator
            array_access_iter  = array_access.begin();
            array_access_iter != array_access.end();
            array_access_iter++) {

          SgPntrArrRefExp *pntr_array_ref = isSgPntrArrRefExp(*array_access_iter);
          SgVarRefExp *array_var_ref = isSgVarRefExp(pntr_array_ref->get_lhs_operand());

          array_var_ref = array_var;

          if(array_var_ref) {

            SgVarRefExp *index_var_ref = isSgVarRefExp(pntr_array_ref->get_rhs_operand());

            if(index_var_ref) {

              if(index_var_ref->get_symbol() == increment_var) {

              } else {

                if(result != Unsupported) {
                  result = WrongArrayIndex;
                }
                continue;
              }
            }
          }
        }

      } else if(risc::tools::is_local_variable(*write_symbol_iter) && is_declared_in_loop(*write_symbol_iter, for_stmt)) {

      } else if(isSgArrayType((*write_symbol_iter)->get_type())) {

      } else {
        result = Unsupported;
      }
    }
  }
  return result;
}

SgVariableSymbol* get_increment_variable(SgExpression *expr)
{
  SgPlusPlusOp *plusplus = isSgPlusPlusOp(expr);
  SgMinusMinusOp *minusminus = isSgMinusMinusOp(expr);
  if(plusplus) {

    return isSgVarRefExp(plusplus->get_operand())->get_symbol();
  }

  if(minusminus) {

    return isSgVarRefExp(minusminus->get_operand())->get_symbol();
  }

  return NULL;
}

SgVariableSymbol* get_conditon_variable(SgExpression *expr)
{
  SgBinaryOp *binary = isSgBinaryOp(expr);
  if(binary) {

    SgLessThanOp *less_than = isSgLessThanOp(binary);
    SgGreaterThanOp *greater_than = isSgGreaterThanOp(binary);
    SgLessOrEqualOp *less_or_equal = isSgLessOrEqualOp(binary);
    SgGreaterOrEqualOp *greater_or_equal = isSgGreaterOrEqualOp(binary);

    if(!less_than && !greater_than && !less_or_equal && !greater_or_equal) {
      return NULL;
    }

    SgVarRefExp *left = isSgVarRefExp(binary->get_lhs_operand());
    SgVarRefExp *right = isSgVarRefExp(binary->get_rhs_operand());
    if(right == NULL) {
      SgArrowExp *arrow = isSgArrowExp(binary->get_rhs_operand());
      if(arrow) {
        right = isSgVarRefExp(arrow->get_rhs_operand());
      }
    }
    SgIntVal *int_val = isSgIntVal(binary->get_rhs_operand());

    if(left && right || left && int_val) {
      return left->get_symbol();
    }
  }

  return NULL;
}

SgVariableSymbol* get_initialized_variable(SgStatement *stmt)
{
  SgExprStatement *expr_stmt = isSgExprStatement(stmt);
  SgVariableDeclaration *var_decl = isSgVariableDeclaration(stmt);

  if(expr_stmt) {
    SgAssignOp *assign = isSgAssignOp(expr_stmt->get_expression());

    if(assign) {

      SgVarRefExp *var_ref = isSgVarRefExp(assign->get_lhs_operand());

      return var_ref->get_symbol();
    }
  }

  if(var_decl) {
    SgInitializedName *init = var_decl->get_decl_item(0);

    if(init) {
      return isSgVariableSymbol(init->search_for_symbol_from_symbol_table());
    }
  }
  return NULL;
}

RISCParseStatus status;
std::ofstream myfile;

int main(int argc, char* argv[])
{

  RISCParseCommandlineOptions(argc, argv, &status);

  if (status.dump)
  {
    myfile.open(status.filename.c_str());
  }


  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
  std::cout << "Output filename: " << status.filename << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "SIMD Compiler starting..."
              << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "SIMD: adding command-line options for frontend..."
              << std::endl;
  }

  std::string LWIncludePath;
  std::string LWLibraryPath;

  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";

  // add -I$SYSTEMC_LW_HOME/include
  risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);

  // add rose:skipfinalCompileStep");
  risc::tools::AddCommandlineOption(argc, argv, "-rose:skipfinalCompileStep");


  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "SIMD: parsing, building ROSE IR..." << std::endl;
  }
  risc::Design d(argc,argv);

  find_simd_loops(d);

  if (status.dump)
  {
    myfile.close();
  }

  return 0;
}


void find_simd_loops(risc::Design &d)
{
  // this must be applied for channels as well
  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    for(risc::ThreadVectorIter
        iter_thread  = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::stringstream ss;
      ss <<  std::endl;


      ss << "Simulation thread: "
         << SageInterface::get_name(isSgMemberFunctionDeclaration((*iter_thread)->get_ast_node()->get_declaration())->get_associatedClassDeclaration())
         << "::"
         << (*iter_thread)->get_name() << std::endl;

      if (status.verbosityLevel >= RISC_VERY_VERBOSE) {

        std::cout << ss.str();

        if(status.dump) {
          output.push_back(ss.str());

          for(std::list<std::string>::iterator
              iter  = output.begin();
              iter != output.end();
              iter++) {
            myfile <<  *iter;
          }

          output.clear();
        }
      }


      std::list<SgNode*> expressions;
      is_simd_compatible((*iter_thread)->get_ast_node()->get_body(), false, expressions);
    }

    for(risc::CThreadVectorIter
        iter_cthread  = (*iter_module)->cthreads_.begin();
        iter_cthread != (*iter_module)->cthreads_.end();
        iter_cthread++) {
    }

    for(risc::MethodVectorIter
        iter_method  = (*iter_module)->methods_.begin();
        iter_method != (*iter_module)->methods_.end();
        iter_method++) {
    }
  }

  std::stringstream ss1;
  ss1 <<  std::endl;
  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {

    std::cout << ss1.str();

    if(status.dump) {
      output.push_back(ss1.str());
    }
  }

  std::set<SgFunctionDefinition*> completed;


  for(std::set<SgFunctionDefinition*>::iterator
      iter  = future_functions.begin();
      iter != future_functions.end();
      iter++) {

    if(completed.find(*iter) == completed.end()) {

      std::stringstream ss;

      if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
        ss << "Start Function: ";
        if(isSgMemberFunctionDeclaration((*iter)->get_declaration())) {

            ss << SageInterface::get_name(isSgMemberFunctionDeclaration((*iter)->get_declaration())->get_associatedClassDeclaration())
                      << "::"
                      << SageInterface::get_name((*iter)->get_declaration())
                      << std::endl;
        } else {

            ss << SageInterface::get_name((*iter)->get_declaration())
                      << std::endl;
        }
        std::cout << ss.str();
      }




      if(status.dump) {
        output.push_back(ss.str());
        ss.str("");
      }

      completed.insert(*iter);

      std::list<SgNode*> expressions;
      SimdCompatibility result = is_simd_compatible((*iter)->get_body(), true, expressions);


      if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
        if (status.dump)
        {
          ss << "  Function: " << SageInterface::get_name((*iter)->get_declaration());
        }
      }


      if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
        if(result == Supported) {

          ss << " is supported." << std::endl << std::endl;;

        } else {

          ss << " is not supported." << std::endl << std::endl;
        }
      }

      if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
        std::cout << ss.str();
      }

      if(status.dump) {
        output.push_back(ss.str());
      }
    }
  }


  if (status.dump) {
    for(std::list<std::string>::iterator
        iter  = output.begin();
        iter != output.end();
        iter++) {
      myfile <<  *iter;
    }
  }
}

SimdCompatibility is_simd_compatible(SgStatement *stmt, bool is_in_loop, std::list<SgNode*> &expressions)
{
  SgBasicBlock *bb = isSgBasicBlock(stmt);
  if(bb) {

    SimdCompatibility status = Supported;

    for(SgStatementPtrList::iterator
        stmt_iter  = bb->get_statements().begin();
        stmt_iter != bb->get_statements().end();
        stmt_iter++) {

      SimdCompatibility temp_status = is_simd_compatible(*stmt_iter, is_in_loop, expressions);

      if(temp_status > status) {
        status = temp_status;
      }
    }
    return status;
  }

  SgSwitchStatement *switch_stmt = isSgSwitchStatement(stmt);
  if(switch_stmt) {

    std::vector<SgStatement*> switch_cases
      = SageInterface::getSwitchCases(switch_stmt);

    for(std::vector<SgStatement*>::iterator
        case_iter  = switch_cases.begin();
        case_iter != switch_cases.end();
        case_iter++) {

      SgCaseOptionStmt *switch_case = isSgCaseOptionStmt(*case_iter);
      SgDefaultOptionStmt *default_case = isSgDefaultOptionStmt(*case_iter);
      SgBasicBlock *bb = NULL;

      if(switch_case) {
        is_simd_compatible(switch_case->get_body(), false, expressions);
      }

      if(default_case) {
        is_simd_compatible(default_case->get_body(), false, expressions);
      }
    }
    return Unsupported;
  }

  SgGotoStatement *goto_stmt = isSgGotoStatement(stmt);
  if(goto_stmt) {
    return Unsupported;
  }

  SgForStatement *for_stmt = isSgForStatement(stmt);
  if(for_stmt) {

    // This checks if the loop has one initialization statement, one condition
    // expression, and one increment statement
    bool satisfies_canonical = SageInterface::isCanonicalForLoop(for_stmt);
    bool satisfies_loop_variables = true;
    bool satisfies_body_analysis = true;
    bool satisfies_write_analysis = true;

    // Here, we check if the loop has the following structure:
    // for(int i = ...; i<...; i++)
    SgVariableSymbol *increment_var = get_increment_variable(for_stmt->get_increment());
    SgVariableSymbol *conition_var = get_conditon_variable(isSgExprStatement(for_stmt->get_test())->get_expression());
    SgVariableSymbol *initialized_var = get_initialized_variable(for_stmt->get_init_stmt().front());

    if(increment_var != conition_var || conition_var != initialized_var) {
      satisfies_loop_variables = false;
    }

    // We are analyzing all the statements in the loop body
    std::list<SgNode*> local_expressions;
    SimdCompatibility body_status = is_simd_compatible(for_stmt->get_loop_body(), true, local_expressions);

    // here we check what is the result of the analysis of the loop body
    if(body_status == Unsupported || body_status == WrongArrayIndex || body_status == VariableConflict) {
      satisfies_body_analysis = false;
    }

    SimdCompatibility write_status = writes_only_simd_compatible_variables(local_expressions, initialized_var, for_stmt);

    if(write_status == Unsupported || write_status == WrongArrayIndex) {
      satisfies_write_analysis = false;
    }

    if(write_status == Supported && body_status == WrongArrayIndex) {
      satisfies_write_analysis = true;
      satisfies_body_analysis = true;
    }

    if(write_status == Supported && body_status == VariableConflict) { // new
      satisfies_write_analysis = true;
      satisfies_body_analysis = true;
    }

    if(satisfies_canonical && satisfies_loop_variables && satisfies_body_analysis && satisfies_write_analysis) {
      std::cout << "  Found loop in line: " << for_stmt->get_file_info()->get_line() << " is qualified." << std::endl;
      if (status.dump)
      {
        myfile << "  Found loop in line: " << for_stmt->get_file_info()->get_line() << " is qualified." << std::endl;
      }

    } else {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "  Found loop in line: " << for_stmt->get_file_info()->get_line() << " is not qualified." << std::endl;
        if (status.dump)
        {
          myfile << "  Found loop in line: " << for_stmt->get_file_info()->get_line() << " is not qualified." << std::endl;
        }
      }
    }

    expressions.insert(expressions.begin(), local_expressions.begin(), local_expressions.end());

    if(satisfies_canonical && satisfies_loop_variables && satisfies_body_analysis && satisfies_write_analysis) {
      return Supported;
    }

    if(body_status == WrongArrayIndex || write_status == WrongArrayIndex) {
      return WrongArrayIndex;
    }

    if(body_status == VariableConflict && write_status == Supported ) {
      return Supported;
    }

    return Unsupported;
  }

  SgExprStatement *expr_stmt = isSgExprStatement(stmt);
  if(expr_stmt) {
    if(is_in_loop) {
      expressions.push_back(expr_stmt);
    }
    return Supported;
  }

  SgIfStmt *if_stmt = isSgIfStmt(stmt);
  if(if_stmt) {

    if(is_in_loop) {
      expressions.push_back(if_stmt->get_conditional());
    }

    SimdCompatibility temp_status = is_simd_compatible(if_stmt->get_true_body(), is_in_loop, expressions);

    if(if_stmt->get_false_body()) {

      SimdCompatibility temp_status2 = is_simd_compatible(if_stmt->get_false_body(), is_in_loop, expressions);

      if(temp_status2 > temp_status)
        temp_status = temp_status2;
    }

    return temp_status;
  }


  SgWhileStmt *while_stmt = isSgWhileStmt(stmt);
  if(while_stmt) {

    std::set<SgVariableSymbol*> read_variables_condition;
    std::set<SgVariableSymbol*> write_variables_condition;

    std::set<SgVariableSymbol*> read_variables_body;
    std::set<SgVariableSymbol*> write_variables_body;

    analyze_statement(while_stmt->get_condition(), read_variables_condition, write_variables_condition);

    std::list<SgNode*> local_expressions;
    SimdCompatibility loop_compability = is_simd_compatible(while_stmt->get_body(), true, local_expressions);

    for(std::list<SgNode*>::iterator
        iter  = local_expressions.begin();
        iter != local_expressions.end();
        iter++) {

      analyze_statement(isSgStatement(*iter), read_variables_body, write_variables_body);
    }

    expressions.insert(expressions.begin(), local_expressions.begin(), local_expressions.end());

    if(loop_compability == Unsupported) {

      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << Unsupported << std::endl;
        if (status.dump)
        {
          myfile << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << Unsupported << std::endl;
        }
      }
      return Unsupported;
    }

    if(writes_only_simd_compatible_variables_while_loop(write_variables_condition, while_stmt) == VariableConflict) {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        if (status.dump)
        {
          myfile << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        }
      }
      return VariableConflict;
    }
    if(writes_only_simd_compatible_variables_while_loop(write_variables_body, while_stmt) == VariableConflict) {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        if (status.dump)
        {
          myfile << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        }
      }
      return VariableConflict;
    }

    std::vector<SgVariableSymbol*> read_body_write_condition;
    std::vector<SgVariableSymbol*> write_body_read_condition;
    std::vector<SgVariableSymbol*> write_body_write_condition;

    std::set_intersection(
        read_variables_body.begin(),
        read_variables_body.end(),
        write_variables_condition.begin(),
        write_variables_condition.end(),
        std::back_inserter(read_body_write_condition));

    std::set_intersection(
        write_variables_body.begin(),
        write_variables_body.end(),
        read_variables_condition.begin(),
        read_variables_condition.end(),
        std::back_inserter(write_body_read_condition));

    std::set_intersection(
        write_variables_body.begin(),
        write_variables_body.end(),
        write_variables_condition.begin(),
        write_variables_condition.end(),
        std::back_inserter(write_body_write_condition));


    if(read_body_write_condition.empty() && write_body_read_condition.empty() && write_body_write_condition.empty()) {

      std::cout << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is qualified." << std::endl;
      if (status.dump)
      {
        myfile << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is qualified." << std::endl;
      }
      return Supported;
    } else {

      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        if (status.dump)
        {
          myfile << "  Found loop in line: " << while_stmt->get_file_info()->get_line() << " is not qualified. Error code: " << VariableConflict << std::endl;
        }
      }
      return VariableConflict;
    }
  }

  SgVariableDeclaration *var_decl = isSgVariableDeclaration(stmt);
  if(var_decl) {

    if(is_in_loop) {
      expressions.push_back(var_decl);
    }
    return Supported;
  }

  if(isSgLabelStatement(stmt)) {
    return Supported;
  }

  if(isSgNullStatement(stmt)) {
    return Supported;
  }

  if(isSgBreakStmt(stmt)) {
    return Supported;
  }

  if(isSgContinueStmt(stmt)) {
    return Supported;
  }

  if(isSgReturnStmt(stmt)) {
    return Supported;
  }

  std::cout << "Error: " << stmt->class_name() << ", " << stmt->unparseToString() << std::endl;
  std::cout << "       Line: " << stmt->get_file_info()->get_line() << std::endl;
  assert(0 && "I should not be here!");
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
