#include "module.h"

#include <algorithm>

#include "rose.h"

#include "design.h"

#include "cthread.h"
#include "event.h"
#include "event.h"
#include "event_and_list.h"
#include "event_or_list.h"
#include "function.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "inoutport.h"
#include "inport.h"
#include "method.h"
#include "module_instance.h"
#include "outport.h"
#include "port.h"
#include "primitive_channel.h"
#include "primitive_channel_instance.h"
#include "thread.h"
#include "variable.h"

#include "../tools/misc.h"
#include "../tools/class_is_derived_from.h"


risc::Module::Module(SgClassDefinition *ast_node):
  Class(ast_node, ModuleType),
  instance_counter_(0)
{
  initialize_functions();
}

risc::Module::Module(SgClassDefinition *ast_node, Type type):
  Class(ast_node, type),
  instance_counter_(0)
{
  initialize_functions();
}

risc::Module::~Module()
{
  for(ThreadVectorIter
      iter  = threads_.begin();
      iter != threads_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(CThreadVectorIter
      iter  = cthreads_.begin();
      iter != cthreads_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(MethodVectorIter
      iter  = methods_.begin();
      iter != methods_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(ResetSignalVectorIter
      iter  = reset_signals_.begin();
      iter != reset_signals_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(ResetSignalVectorIter
      iter  = async_reset_signals_.begin();
      iter != async_reset_signals_.end();
      iter++) {
    delete *iter;
    *iter = NULL;
  }
}

void risc::Module::initialize_functions()
{
  // Traversing for functions and getting all associated SystemC semantics is a
  // little bit tricky. We have to distinguish between a function in the
  // classic C++ style and a SystemC process which can have a static
  // sensitivity list. We can obtain these information in the ctor of the
  // individual module. This is the concept:
  //
  // (1) We traverse for all function definitions in the class definition
  //     and put them in a temporary container
  // (2) We have to analyze the ctor of the module and figure out which
  //     functions are registrated as Thread, CThread, and Method.
  // (3) We pick these function from the temporary container and put them
  //     into the corresponding process lists for threads, cthreads, and
  //     methods
  // (4) We analyze the following static sensitivity list (detials later)
  //     and associate the sensitivity list with the SystemC processes
  // (5) Finally, only 'pure' C++ function are in the temporary container

  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(get_ast_node()->get_declaration());

  // sometimes getDefaultConstructor provides a non defining declaration so
  // we have to cover this case
  ctor = isSgMemberFunctionDeclaration(ctor->get_definingDeclaration());

  // get dtor of the class (module)
  SgMemberFunctionDeclaration *dtor =
    SageInterface::getDefaultDestructor(get_ast_node()->get_declaration());

  // sometimes getDefaultConstructor provides a non defining declaration so
  // we have to cover this case
  if(dtor) {
    dtor = isSgMemberFunctionDeclaration(dtor->get_definingDeclaration());
  }

  // Find all functions in the class defintion
  std::vector<SgNode*> function_definitions =
    NodeQuery::querySubTree(get_ast_node(), V_SgFunctionDefinition);

  // The constructor is in the function_definitions container because rose
  // treats a constructor as a special kind of a function. So, we have to
  // search for the ctor and delete it.
  if(ctor) {

    function_definitions.erase(
      std::remove(function_definitions.begin(), function_definitions.end(),
               ctor->get_definition()),
      function_definitions.end());
  }

  // same for the destructor
  if(dtor) {
    function_definitions.erase(
      std::remove(function_definitions.begin(), function_definitions.end(),
               dtor->get_definition()),
      function_definitions.end());
  }

  if(ctor) {

    // Now, we will decide if each individal function has a specific SystemC
    // semantic (like a THREAD, ...). Please consider how a SystemC module
    // constructor is transformed after the preprocessing below:

    /* INPUT of a SystemC ctor before preprocessing
     * Foo(sc_module_name mn):
     *   sc_module(mn)
     * {
     *   SC_THREAD(my_thread1);
     *
     *   SC_THREAD(my_thread2);
     *   sensitive << e1 << e2;
     * }
     *
     * OUTPUT of a SystemC ctor after preprocessing
     * inline Foo(class sc_module_name mn): sc_module(mn) {
     *   {
     *     class sc_process_handle my_thread1_handle =
     *       sc_core::sc_get_curr_simcontext()->create_thread_process(
     *         "my_thread1", false, &Foo::my_thread1, (this), 0);
     *     (this)->sensitive << my_thread1_handle;
     *     (this)->sensitive_pos << my_thread1_handle;
     *     (this)->sensitive_neg << my_thread1_handle;
     *   };
     *   {
     *     class sc_process_handle my_thread2_handle =
     *       sc_core::sc_get_curr_simcontext()->create_thread_process(
     *           "my_thread2", false, &Foo::my_thread2, (this), 0);
     *   (this)->sensitive << my_thread2_handle;
     *   (this)->sensitive_pos << my_thread2_handle;
     *   (this)->sensitive_neg << my_thread2_handle;
     *   };
     *   ((this)->sensitive << ((this)->e1)) << ((this)->e2);
     * }
     */
    // This example shows that each process registration is transformed in a
    // dedicated basic block. The sensitivity list is immediately after the
    // basic block.

    // First, we search for basic blocks in the ctor. In the basic block we
    // might expect the registration of the function to the sysc kernel.
    // For this reason, we check if the specific sub basic block in the ctor
    // includes a function call to create_XXX_process. If this is the case, we
    // are dealing with a function registration.
    // Afterwards, we have to search for event registrations. An event
    // registration happens immeditly after the basic block. So, we search for
    for(SgStatementPtrList::iterator
        iter  = ctor->get_definition()->get_body()->get_statements().begin();
        iter != ctor->get_definition()->get_body()->get_statements().end();
        iter++) {

      SgBasicBlock *bb = isSgBasicBlock(*iter);

      if(bb) {

        std::vector<SgNode*> function_calls =
          NodeQuery::querySubTree(bb, V_SgFunctionCallExp);

        bool has_potential_sensitivity_list = false;
        bool has_potential_reset_list = false;

        for(std::vector<SgNode*>::iterator
            iter_call  = function_calls.begin();
            iter_call != function_calls.end();
            iter_call++) {

          SgFunctionCallExp *call_exp = isSgFunctionCallExp(*iter_call);
          SgFunctionDeclaration *fd =
            call_exp->getAssociatedFunctionDeclaration();

          SgFunctionDefinition *process_def = NULL;
          Thread *new_thread = NULL;
          Method *new_method = NULL;
          CThread *new_cthread = NULL;

          if(fd->get_name() == "create_thread_process") {

            SgClassDeclaration *class_decl = get_ast_node()->get_declaration();
            SgClassDefinition *class_def
              = isSgClassDeclaration(
                    class_decl->get_definingDeclaration())->get_definition();

            SgExprListExp *expr_list = call_exp->get_args();

            process_def =
              SageInterface::findFunctionDeclaration(
                  class_def,
                  isSgStringVal(((expr_list->get_expressions())[0]))->get_value(),
                  NULL,
                  true)->
              get_definition();

            has_potential_sensitivity_list = true;
            new_thread = new Thread(process_def, this);
            threads_.push_back(new_thread);
  /************** Added by Guantao Liu 08/28/2015 **************/
            new_thread->set_starting_node(call_exp);
  /************** Added by Guantao Liu 08/28/2015 **************/
          }

          if(fd->get_name() == "create_cthread_process") {

            SgClassDeclaration *class_decl = get_ast_node()->get_declaration();
            SgClassDefinition *class_def
              = isSgClassDeclaration(
                    class_decl->get_definingDeclaration())->get_definition();

            SgExprListExp* expr_list = call_exp->get_args();

            process_def =
              SageInterface::findFunctionDeclaration(
                  class_def,
                  isSgStringVal(((expr_list->get_expressions())[0]))->get_value(),
                  NULL,
                  true)->
              get_definition();

            new_cthread = new CThread(process_def, this);
            cthreads_.push_back(new_cthread);
  /************** Added by Guantao Liu 08/28/2015 **************/
            new_cthread->set_starting_node(call_exp);
  /************** Added by Guantao Liu 08/28/2015 **************/

            // Searching for the associated clock
            std::vector<SgNode*> events =
              NodeQuery::querySubTree(*iter, V_SgVarRefExp);

            for(std::vector<SgNode*>::iterator
                event_iter  = events.begin();
                event_iter != events.end(); event_iter++) {

              SgVarRefExp* var_ref = isSgVarRefExp(*event_iter);
              SgVariableDefinition* var_def =
                isSgVariableDeclaration(
                    var_ref->get_symbol()->get_declaration()->
                        get_declaration())->
                get_definition();

              // We are looking for the associated class definition of the
              // var reference
              SgClassDefinition *class_def =
                isSgClassDeclaration(
                    isSgClassDeclaration(
                        isSgClassDeclaration(
                            isSgClassType(
                              var_ref->get_type())->get_declaration()))->
                    get_definingDeclaration())->
                get_definition();

              // We are looking for all variable references with the type of
              // an event
              if(risc::tools::is_sc_in_clk(class_def)) {

                new_cthread->clk_port_ = (InPort*)find_port(var_def);
              } else {

                if(risc::tools::is_sc_in_bool(class_def)) {
                  new_cthread->clk_port_ = (InPort*)find_port(var_def);
                }
              }
            }
            has_potential_reset_list = true;
          }

          if(fd->get_name() == "create_method_process") {

            SgExprListExp* expr_list = call_exp->get_args();

            std::vector<SgNode*> func_decls =
              NodeQuery::querySubTree(
                  SageInterface::getEnclosingClassDefinition(expr_list),
                  V_SgFunctionDefinition);

            for(std::vector<SgNode*>::iterator
                iter  = func_decls.begin();
                iter != func_decls.end();
                iter++){
              SgFunctionDefinition *func_def = isSgFunctionDefinition(*iter);
            }

            SgClassDeclaration *class_decl = get_ast_node()->get_declaration();
            SgClassDefinition *class_def
              = isSgClassDeclaration(
                    class_decl->get_definingDeclaration())->get_definition();

            process_def =
              SageInterface::findFunctionDeclaration(
                  class_def,
                  isSgStringVal(((expr_list->get_expressions())[0]))
                    ->get_value(),
                  NULL,
                  true)->
              get_definition();

            has_potential_sensitivity_list  = true;
            new_method = new Method(process_def, this);
            methods_.push_back(new_method);
  /************** Added by Guantao Liu 08/28/2015 **************/
            new_method->set_starting_node(call_exp);
  /************** Added by Guantao Liu 08/28/2015 **************/
          }

          // Delete the process in the temporary function list
          if(process_def) {
            function_definitions.erase(
                                     std::remove(function_definitions.begin(),
                                              function_definitions.end(),
                                              process_def),
                                     function_definitions.end());
          }

          // Now, we search for a static sensitivity list
          // The static sensitivity list should follow immediately after the
          // registration of the function. General idea: We analyze statement
          // for statement until the next statement is a basic block
          if(has_potential_sensitivity_list) {
            SgStatementPtrList::iterator previous_iter = iter;

            // get next statement in the constructor
            iter++;

            // In this loop, we iterate the constructor body statements until we
            // reach the end of the constructor body or the next statement is a
            // new basic block
            while(iter !=
                ctor->get_definition()->get_body()->get_statements().end()) {

              SgExprStatement *expr_stmt = isSgExprStatement(*iter);
              if(expr_stmt) {

                SgFunctionCallExp *func_call =
                  isSgFunctionCallExp(expr_stmt->get_expression());
                if(func_call) {


                  // TODO: I should compare against the symbol to catch the
                  //       right function.

                  // Find all variable references in the current statement.
                  // In the statement itself, we will look for references with the
                  // type of an event
                  std::vector<SgNode*> events
                    = NodeQuery::querySubTree(*iter, V_SgVarRefExp);

                  for(std::vector<SgNode*>::iterator
                      event_iter  = events.begin();
                      event_iter != events.end();
                      event_iter++) {

                    SgVarRefExp* var_ref = isSgVarRefExp(*event_iter);
                    SgVariableDefinition* var_def =
                      isSgVariableDeclaration(
                          var_ref->get_symbol()->get_declaration()->
                              get_declaration())->get_definition();

                    // We are looking for all variable references with the type of
                    // an event
                    if(risc::tools::is_sc_event(var_ref)) {

                      Event *event = find_event(var_def);

                      if(new_thread)
                        new_thread->static_event_sensitivity_list_.push_back(
                                                                       event);
                      else
                        new_method->static_event_sensitivity_list_.push_back(
                                                                       event);
                    }


                    // We are looking for all variable references with the type of
                    // a port_base
                    if(risc::tools::is_derived_from_sc_port_base(var_ref)) {

                      Port *port = find_port(var_def);

                      if(new_thread)
                        new_thread->static_port_sensitivity_list_.push_back(port);
                      else
                        new_method->static_port_sensitivity_list_.push_back(port);
                    }
                  }
                }

                previous_iter = iter;
                iter++;

              } else {
                break;
              }
            }

            iter = previous_iter;
          }


          if(has_potential_reset_list)  {

            SgStatementPtrList::iterator previous_iter = iter;

            // get next statement in the constructor
            iter++;

            // In this loop, we iterate the constructor body statements until we
            // reach the end of the constructor body or the next statement is a
            // new basic block
            while(iter !=
                ctor->get_definition()->get_body()->get_statements().end()) {

              SgExprStatement *expr_stmt = isSgExprStatement(*iter);
              if(expr_stmt) {

                SgFunctionCallExp *func_call =
                  isSgFunctionCallExp(expr_stmt->get_expression());

                if(func_call) {

                  // TODO: I should compare against the symbol to catch the
                  //       right function.
                  SgFunctionDeclaration *func_decl =
                    func_call->getAssociatedFunctionDeclaration();

                  Port *port = NULL;

                  if(SageInterface::get_name(func_decl) == "reset_signal_is" ||
                      SageInterface::get_name(func_decl) ==
                      "async_reset_signal_is") {

                    SgExprListExp *arguments = func_call->get_args();

                    SgExpressionPtrList expressions =
                      arguments->get_expressions();

                    SgVarRefExp *var_ref =
                      isSgVarRefExp(isSgArrowExp(
                          expressions.at(0))->get_rhs_operand());
                    SgBoolValExp *bool_val = isSgBoolValExp(expressions.at(1));

                    SgVariableDefinition* var_def = isSgVariableDeclaration(
                        var_ref->get_symbol()->get_declaration()->
                            get_declaration())->get_definition();

                    port = find_port(var_def);

                    if(SageInterface::get_name(func_decl) == "reset_signal_is") {

                      ResetSignal *registered_reset_signal = NULL;

                      // Protection: Avoid same reset signal twice
                      //             Set cannot compare because its a pointer
                      for(ResetSignalVector::iterator
                          iter_reset_signals  = reset_signals_.begin();
                          iter_reset_signals != reset_signals_.end();
                          iter_reset_signals++) {

                        if(((*iter_reset_signals)->first == port) &&
                            (*iter_reset_signals)->second->get_value() ==
                            bool_val->get_value()) {
                          registered_reset_signal = *iter_reset_signals;
                        }
                      }

                      if(!registered_reset_signal) {
                        ResetSignal *reset_signal = new ResetSignal(port,
                                                            bool_val);
                        new_cthread->reset_signal_.push_back(reset_signal);
                        reset_signals_.push_back(reset_signal);
                      } else {
                        new_cthread->reset_signal_.push_back(
                                                       registered_reset_signal);
                      }
                    }

                    if(SageInterface::get_name(func_decl) ==
                        "async_reset_signal_is") {

                      ResetSignal *registered_async_reset_signal = NULL;

                      // Protection: Avoid same reset signal twice
                      //             Set cannot compare because its a pointer
                      for(ResetSignalVector::iterator
                          iter_reset_signals  = async_reset_signals_.begin();
                          iter_reset_signals != async_reset_signals_.end();
                          iter_reset_signals++) {

                        if(((*iter_reset_signals)->first == port) &&
                            (*iter_reset_signals)->second->get_value() ==
                            bool_val->get_value()) {
                          registered_async_reset_signal = *iter_reset_signals;
                        }
                      }

                      if(!registered_async_reset_signal) {
                        ResetSignal *async_reset_signal =
                          new ResetSignal(port, bool_val);
                        new_cthread->async_reset_signal_.push_back(
                                                             async_reset_signal);
                        async_reset_signals_.push_back(async_reset_signal);
                      } else {
                        new_cthread->async_reset_signal_.
                            push_back(registered_async_reset_signal);
                      }
                    }
                  }
                }

                previous_iter = iter;
                iter++;

              } else {
                break;
              }
            }

            iter = previous_iter;
          }
        }
      }
    }
  }

  // All other functions are classic C++ functions
  for(std::vector<SgNode*>::iterator
      iter  = function_definitions.begin();
      iter != function_definitions.end();
      iter++) {

    functions_.push_back(
        new Function(isSgFunctionDefinition(*iter),FunctionType, this));
  }
}


void risc::Module::print_module_definition()
{
  std::cout << "Definition of " << get_ast_type_name() << std::endl;

  // print module variables
  if(sub_modules_.size() != 0) {

    std::cout << " Submodules" << std::endl;
    for(ModuleInstanceVectorConstIter
       iter  = sub_modules_.begin();
       iter != sub_modules_.end();
       iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print builtin variables
  if(variables_.size() != 0) {

    std::cout << " Variables" << std::endl;
    for(VariableVectorConstIter
        iter  = variables_.begin();
        iter != variables_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print events
  if(events_.size() != 0) {

    std::cout << " Events" << std::endl;
    for(EventVectorConstIter
        iter  = events_.begin();
        iter != events_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print and_event_lists
  if(event_and_lists_.size() != 0) {

    std::cout << " Event and lists:" << std::endl;
    for(EventAndListVectorIter
        iter  = event_and_lists_.begin();
        iter != event_and_lists_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print or_event_lists
  if(event_or_lists_.size() != 0) {

    std::cout << " Event or lists" << std::endl;
    for(EventOrListVectorIter
        iter  = event_or_lists_.begin();
        iter != event_or_lists_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print ports
  if( (inports_.size() != 0) ||
      (outports_.size() != 0) ||
      (inoutports_.size() != 0)) {

    std::cout << " Ports" << std::endl;

    if(!inports_.empty())
      std::cout << "  In" << std::endl;
    for(InPortVectorConstIter
        iter  = inports_.begin();
        iter != inports_.end();
        iter++) {

      std::cout << "   " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;

      std::cout << "   mapped channel:";
      std::set<risc::Object*> mapped_channels;
      (*iter)->get_mapped_channels(mapped_channels);
      for(std::set<risc::Object*>::iterator
          channel_iter  = mapped_channels.begin();
          channel_iter != mapped_channels.end();
          channel_iter++) {

        std::cout << " " << (*channel_iter)->get_name();
      }
      std::cout << std::endl;

      std::cout << "   mapped channel types:";
      std::set<risc::Class*> mapped_channel_types;
      (*iter)->get_mapped_channel_types(mapped_channel_types);
      for(std::set<risc::Class*>::iterator
          type_iter  = mapped_channel_types.begin();
          type_iter != mapped_channel_types.end();
          type_iter++) {

        std::cout << " " << (*type_iter)->get_name();
      }
      std::cout << std::endl;
    }

    if(!outports_.empty())
      std::cout << "  Out" << std::endl;
    for(OutPortVectorConstIter
        iter  = outports_.begin();
        iter != outports_.end();
        iter++) {

      std::cout << "   " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;

      std::cout << "   mapped channel:";
      std::set<risc::Object*> mapped_channels;
      (*iter)->get_mapped_channels(mapped_channels);
      for(std::set<risc::Object*>::iterator
          channel_iter  = mapped_channels.begin();
          channel_iter != mapped_channels.end();
          channel_iter++) {

        std::cout << " " << (*channel_iter)->get_name();
      }
      std::cout << std::endl;

      std::cout << "   mapped channel types:";
      std::set<risc::Class*> mapped_channel_types;
      (*iter)->get_mapped_channel_types(mapped_channel_types);
      for(std::set<risc::Class*>::iterator
          type_iter  = mapped_channel_types.begin();
          type_iter != mapped_channel_types.end();
          type_iter++) {

        std::cout << " " << (*type_iter)->get_name();
      }
      std::cout << std::endl;
    }

    if(!inoutports_.empty())
      std::cout << "  InOut" << std::endl;
    for(InOutPortVectorConstIter
        iter  = inoutports_.begin();
        iter != inoutports_.end();
        iter++) {

      std::cout << "   " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;

      std::cout << "   mapped channel:";
      std::set<risc::Object*> mapped_channels;
      (*iter)->get_mapped_channels(mapped_channels);
      for(std::set<risc::Object*>::iterator
          channel_iter  = mapped_channels.begin();
          channel_iter != mapped_channels.end();
          channel_iter++) {

        std::cout << " " << (*channel_iter)->get_name();
      }
      std::cout << std::endl;

      std::cout << "   mapped channel types:";
      std::set<risc::Class*> mapped_channel_types;
      (*iter)->get_mapped_channel_types(mapped_channel_types);
      for(std::set<risc::Class*>::iterator
          type_iter  = mapped_channel_types.begin();
          type_iter != mapped_channel_types.end();
          type_iter++) {

        std::cout << " " << (*type_iter)->get_name();
      }
      std::cout << std::endl;
    }
  }

  // print primitive channels
  if(sub_prim_channel_.size() != 0) {
    std::cout << " Primitive Channels" << std::endl;

    for(PrimitiveChannelInstanceVectorConstIter
        iter  = sub_prim_channel_.begin();
        iter != sub_prim_channel_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print hierarchical channels
  if(sub_hier_channel_.size() != 0) {

    std::cout << " Hierarchical Channels " << std::endl;
    for(HierarchicalChannelInstanceVectorConstIter
        iter  = sub_hier_channel_.begin();
        iter != sub_hier_channel_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print member functions
  if(functions_.size() != 0) {

    std::cout << " Member Functions (type, name)" << std::endl;
    for(FunctionVectorConstIter
        iter  = functions_.begin();
        iter != functions_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print Threads
  if(threads_.size() != 0) {

    std::cout << " Threads (name: static sensitivity list)" << std::endl;
    for(ThreadVectorConstIter
        iter  = threads_.begin();
        iter != threads_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_name() << ": ";
      (*iter)->print_event_list();
    }
  }

  // print CThreads
  if(cthreads_.size() != 0) {

    std::cout << " CThreads (name, clock)" << std::endl;
    for(CThreadVectorConstIter
        iter  = cthreads_.begin();
        iter != cthreads_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_name()
                << ", " << (*iter)->clk_port_->get_name() << std::endl;

      if((*iter)->reset_signal_.size() != 0 ) {

        std::cout << "    " << "Reset signals: ";
        for(ResetSignalVector::iterator
            iter_reset_signal  = (*iter)->reset_signal_.begin();
            iter_reset_signal != (*iter)->reset_signal_.end();
            iter_reset_signal++) {

          std::cout << "("  << (*iter_reset_signal)->first->get_name()
                    << ", " << (*iter_reset_signal)->second->get_value()
                    << ") ";
        }
        std::cout << std::endl;
      }

      if((*iter)->async_reset_signal_.size() != 0 ) {

        std::cout << "    " << "Reset signals: ";
        for(ResetSignalVector::iterator
            iter_async_reset_signal  = (*iter)->async_reset_signal_.begin();
            iter_async_reset_signal != (*iter)->async_reset_signal_.end();
            iter_async_reset_signal++) {

          std::cout << "("  << (*iter_async_reset_signal)->first->get_name()
                    << ", " << (*iter_async_reset_signal)->second->get_value()
                    << ") ";
        }
        std::cout << std::endl;
      }
    }
  }

  // print Methods functions
  if(methods_.size() != 0) {

    std::cout << " Methods (name: static sensitivity list)" << std::endl;
    for(MethodVectorConstIter
        iter  = methods_.begin();
        iter != methods_.end();
        iter++) {

      std::cout << "  " << (*iter)->get_name() << ": ";
      (*iter)->print_event_list();
    }
  }

  // print Methods functions
  if(reset_signals_.size() != 0) {

    std::cout << " Reset Signals" << std::endl;
    for(ResetSignalVector::iterator
        iter  = reset_signals_.begin();
        iter != reset_signals_.end();
        iter++) {

      std::cout << "  " << (*iter)->first->get_name() << std::endl;
    }
  }

  if(async_reset_signals_.size() != 0) {

    std::cout << " Async Reset Signals" << std::endl;
    for(ResetSignalVector::iterator
        iter  = async_reset_signals_.begin();
        iter != async_reset_signals_.end();
        iter++) {

      std::cout << "  " << (*iter)->first->get_name() << std::endl;
    }
  }
}

void risc::Module::print_hierarchy(unsigned indent_level)
{
  for(ModuleInstanceVectorConstIter
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {

    for(unsigned i=0; i<indent_level; i++) {
      std::cout << "  ";
    }

    std::cout << (*iter)->get_ast_type_name()
              << " " << (*iter)->get_name() << std::endl;

    for(PrimitiveChannelInstanceVector::iterator
        prim_iter  = (*iter)->get_definition()->sub_prim_channel_.begin();
        prim_iter != (*iter)->get_definition()->sub_prim_channel_.end();
        prim_iter++) {

      for(unsigned i=0; i<indent_level; i++) {
        std::cout << "  ";
      }
      std::cout << (*prim_iter)->get_name() << std::endl;
    }

    for(HierarchicalChannelInstanceVector::iterator
        hier_iter  = (*iter)->get_definition()->sub_hier_channel_.begin();
        hier_iter != (*iter)->get_definition()->sub_hier_channel_.end();
        hier_iter++) {

      for(unsigned i=0; i<indent_level; i++) {
        std::cout << "  ";
      }
      std::cout << (*hier_iter)->get_name() << std::endl;
    }

    (*iter)->get_definition()->print_hierarchy(indent_level+1);
  }

  for(HierarchicalChannelInstanceVectorConstIter
      iter  = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end();
      iter++) {

    for(unsigned i=0; i<indent_level; i++) {
      std::cout << "  ";
    }

    std::cout << (*iter)->get_ast_type_name()
              << " " << (*iter)->get_name() << std::endl;

    for(PrimitiveChannelInstanceVector::iterator
        prim_iter  = (*iter)->get_definition()->sub_prim_channel_.begin();
        prim_iter != (*iter)->get_definition()->sub_prim_channel_.end();
        prim_iter++) {

      for(unsigned i=0; i<indent_level; i++) {
        std::cout << "  ";
      }
      std::cout << (*prim_iter)->get_name() << std::endl;
    }

    for(HierarchicalChannelInstanceVector::iterator
        hier_iter  = (*iter)->get_definition()->sub_hier_channel_.begin();
        hier_iter != (*iter)->get_definition()->sub_hier_channel_.end();
        hier_iter++) {

      for(unsigned i=0; i<indent_level; i++) {
        std::cout << "  ";
      }
      std::cout << (*hier_iter)->get_name() << std::endl;
    }

    (*iter)->get_definition()->print_hierarchy(indent_level+1);
  }
}


std::string risc::Module::get_qualified_type_name()
{
  std::string qualified_name
    = get_ast_node()->get_declaration()->get_qualified_name();
  
  return qualified_name;
}

std::string risc::Module::get_mangled_type_name()
{
  std::string mangled_name
    = get_ast_node()->get_declaration()->get_type()->get_mangled();
  tools::find_and_replace_all(mangled_name, "__scope__", "::");
  return mangled_name;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
