#include "data_conflict_table.h"

#include <fstream>

#include "mapped_variable.h"
#include "segment.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/static_path_instance_mapper.h"

#include "../internal_representation/object.h"


int get_id(risc::InstanceTree path, SgVariableSymbol *symbol);

extern DynamicModule *ast_root;

risc::sg::DataConflictTable::DataConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  // here we perform a read/write analysis on each segment in the graph
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    segment_graph_.graph_[*vp.first].analyze_segment();
  }

  determine_conflict_table();
}

void
risc::sg::DataConflictTable::determine_conflict_table()
{
  bool dynamic_analysis = path_instance_mapper_->perform_dynamic_analysis();
  for(int row=0; row < size_of_conflict_table_; row++) {

    // Here, we are getting the segment id and instance id with the associated
    // row index 'row'.
    std::pair<int, int> row_id_and_instance =
      index_to_segment_and_instance_id(row);

    int row_seg = row_id_and_instance.first;
    int row_inst = row_id_and_instance.second;

    for(int column=0; column <= row; column++) {

      // Here, we are getting the segment id and instance id with the associated
      // column index 'column'.
      std::pair<int, int> column_id_and_instance =
        index_to_segment_and_instance_id(column);

      int column_seg = column_id_and_instance.first;
      int column_inst = column_id_and_instance.second;

      if(column_seg == row_seg && row_inst == column_inst) {

        // If only read access in this segment, there will be no conflict
        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)].
            write_variables_.empty()) {

          conflict_table_[row][column] = false;
        } else {

          bool has_conflict = false;

          for(std::list<SymbolWithPath>::iterator
              write_iter  = segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.begin();
              write_iter != segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.end();
              write_iter++) {

            MappedVariableList mapped_variables
              = get_mapped_variable(write_iter->symbol_, row_inst,
                  dynamic_analysis);

            for(MappedVariableList::iterator
                mapped_variable  = mapped_variables.begin();
                mapped_variable != mapped_variables.end();
                mapped_variable++) {

              SgVariableSymbol *variable_symbol = (*mapped_variable).symbol_;

              if(is_conflict_free_variable(variable_symbol)) {
                continue;
              }

              if((tools::is_member_variable(variable_symbol)) ||

                  (tools::is_local_variable(variable_symbol) &&
                  tools::is_static_variable(variable_symbol)) ||

                  tools::is_global_variable(variable_symbol) ||
                  tools::is_namespace_variable(variable_symbol)) {

                if(tools::is_channel_variable((*mapped_variable).symbol_)
                    && !tools::is_reference_variable(write_iter->symbol_)) {

                  mapped_variable->instance_id_
                    = get_channel_id(*write_iter, row_inst, dynamic_analysis);
                }

                has_conflict = true;

                if(debugging_ && !dynamic_analysis) {

                  conflicts_per_instance_[
                    row_inst * max_instances_ + column_inst]++;
                }

                if(debugging_) {

                  // We are displaying the mapped variable instead of the
                  // regular variable ... just in case of a reference
                  Conflict conflict(*mapped_variable);
                  conflict_info_[column][row].conflict_variables_.
                    insert(conflict);
                  conflict_info_[row][column].conflict_variables_.
                    insert(conflict);
                }
              }
            }
          }

          conflict_table_[row][column] = has_conflict;
        }

        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
                .calls_conflict_function_
            || segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
                .deref_pointer_write_){

          if(debugging_) {
            conflict_info_[row][column].conflict_variables_ =
              segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
                conflicts_;
          }
        }

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
            .calls_conflict_function_;

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
            .deref_pointer_write_;

        // we don't need a read check because we have
        // he same segment and instance id
      } else {

        // Here we perform the conflict analysis of the individual parts
        conflict_table_[row][column] = data_conflict_between(
            segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)],
            segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)],
            row,
            column,
            row_inst,
            column_inst,
            dynamic_analysis);
        conflict_table_[column][row] = conflict_table_[row][column];

        if(debugging_ && !dynamic_analysis) {
          if(conflict_table_[row][column]) {
            conflicts_per_instance_[
              row_inst * max_instances_ + column_inst]++;
            conflicts_per_instance_[
              column_inst * max_instances_ + row_inst]++;
          }
        }
      }
    }
  }
}

void
risc::sg::DataConflictTable::color_members_of_class(
    const MappedVariable &variable,
    DataConflictAstAttributes::Status color)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  MappedVariable temp;

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        if(isSgClassType(var_decl->get_decl_item(0)->get_type())) {

          temp.symbol_ = var_symbol;
          temp.instance_id_ = variable.instance_id_;
          color_members_of_class(temp, color);
        }

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {
          if(color == DataConflictAstAttributes::UNUSED) {
            data_conflict_attribute->coloring_read_.clear();
            data_conflict_attribute->coloring_write_.clear();
          } else {
            if(color == DataConflictAstAttributes::READ) {
              data_conflict_attribute->coloring_read_.insert(variable.instance_id_);
            } else {
              data_conflict_attribute->coloring_write_.insert(variable.instance_id_);
            }
          }
        }
      }
    }
  }
}

bool
risc::sg::DataConflictTable::check_members_of_class(
    const MappedVariable &mapped_variable,
    DataConflictAstAttributes::Status color,
    std::set<Conflict> &conflicts)
{
  SgInitializedName *var_decl_stmt = mapped_variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return false;
  }

  // we get the type of the variable
  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return false;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  bool result = false;

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);

    if(var_decl) {
      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        //color_members_of_class
        if(isSgClassType(var_decl->get_decl_item(0)->get_type())) {

          MappedVariable temp(var_symbol, mapped_variable.instance_id_);
          if(check_members_of_class(temp, color, conflicts))  {

            result = true;
          }
        }

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {
          if(color == DataConflictAstAttributes::READ) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_read_.begin();
                iter != data_conflict_attribute->coloring_read_.end();
                iter++) {

              if(*iter == mapped_variable.instance_id_) {
                if(debugging_) {
                  conflicts.insert(MappedVariable(var_symbol,
                        mapped_variable.instance_id_));
                }
                result = true;
              }
            }
          }

          if(color == DataConflictAstAttributes::WRITE) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_write_.begin();
                iter != data_conflict_attribute->coloring_write_.end();
                iter++) {

              if(*iter == mapped_variable.instance_id_) {
                if(debugging_) {
                  conflicts.insert(MappedVariable(var_symbol,
                        mapped_variable.instance_id_));
                }
                result = true;
              }
            }
          }
        }
      }
    }
  }
  return result;
}

bool
risc::sg::DataConflictTable::data_conflict_between(
    Segment &first_seg,
    Segment &second_seg,
    int row, int column,
    int inst_id_first, int inst_id_second,
    bool dynamic_analysis)
{
  if(first_seg.calls_conflict_function_
      || second_seg.calls_conflict_function_
      || first_seg.deref_pointer_write_
      || second_seg.deref_pointer_write_
      || (first_seg.deref_pointer_read_
           && !second_seg.write_variables_.empty())
      || (second_seg.deref_pointer_read_
           && !first_seg.write_variables_.empty())) {

    if(first_seg.calls_conflict_function_
        || first_seg.deref_pointer_write_
        || (first_seg.deref_pointer_read_
            && !second_seg.write_variables_.empty())) {

      if(debugging_) {
        conflict_info_[row][column].conflict_variables_.insert(
            first_seg.conflicts_.begin(),
            first_seg.conflicts_.end());

        conflict_info_[column][row].conflict_variables_.insert(
            first_seg.conflicts_.begin(),
            first_seg.conflicts_.end());
      }
    }

    if(second_seg.calls_conflict_function_
        || second_seg.deref_pointer_write_
        || (second_seg.deref_pointer_read_
             && !first_seg.write_variables_.empty())) {

      if(debugging_) {
        conflict_info_[row][column].conflict_variables_.insert(
            second_seg.conflicts_.begin(),
            second_seg.conflicts_.end());

        conflict_info_[column][row].conflict_variables_.insert(
            second_seg.conflicts_.begin(),
            second_seg.conflicts_.end());
      }
    }

    return true;
  }

  std::set<Conflict> conflicts;
  risc::Design *d = risc::get_design();

  bool conflict = false;

  MappedVariable temp;

  for(std::list<SymbolWithPath>::iterator
      read_iter  = first_seg.read_variables_.begin(),
      read_end   = first_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if(!tools::is_reference_variable(read_iter->symbol_)) {

        temp.symbol_ = read_iter->symbol_;
        temp.instance_id_ = inst_id_first;

        if(tools::is_stack_variable(temp.symbol_)) {
          continue;
        }

        if(is_conflict_free_variable(temp.symbol_)) {
          continue;
        }

        if(tools::is_channel_variable(temp.symbol_)) {
          temp.instance_id_
            = get_channel_id(*read_iter, temp.instance_id_, dynamic_analysis);
        }

        color_symbol(temp, DataConflictAstAttributes::READ);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_first,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {

        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(tools::is_channel_variable(read_iter->symbol_)) {

          int channel_id
            = get_channel_id(*read_iter, inst_id_first, dynamic_analysis);

          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->port_mapping_cache_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  read_iter->symbol_));

          // iterating over all mapped modules
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {

            int sybmol_instance
              = get_id(iter->second, (*mapped_variable).symbol_);

            if (sybmol_instance != -2) {

              mapped_variable->instance_id_ = sybmol_instance;
              color_symbol(*mapped_variable, DataConflictAstAttributes::READ);
            }
          }
        } else {

          color_symbol(*mapped_variable, DataConflictAstAttributes::READ);
        }
      }
    }
  }

  for(std::list<SymbolWithPath>::iterator
      write_iter  = first_seg.write_variables_.begin(),
      write_end   = first_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {

    if(!tools::is_reference_variable(write_iter->symbol_)) {

        temp.symbol_ = write_iter->symbol_;
        temp.instance_id_ = inst_id_first;

        if(tools::is_stack_variable(temp.symbol_)) {
          continue;
        }

        if(is_conflict_free_variable(temp.symbol_)) {
          continue;
        }

        if(tools::is_channel_variable(temp.symbol_)) {
          temp.instance_id_
            = get_channel_id(*write_iter, inst_id_first, dynamic_analysis);
        }

        color_symbol(temp, DataConflictAstAttributes::WRITE);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_first,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {

        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(tools::is_channel_variable(write_iter->symbol_)) {

          int channel_id
            = get_channel_id(*write_iter, inst_id_first, dynamic_analysis);

          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->port_mapping_cache_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  write_iter->symbol_));

          // iterating over all mapped modules
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {

            int sybmol_instance
              = get_id(iter->second, (*mapped_variable).symbol_);

            if (sybmol_instance != -2) {

              mapped_variable->instance_id_ = sybmol_instance;
              color_symbol(*mapped_variable, DataConflictAstAttributes::WRITE);
            }
          }
        } else {

          color_symbol(*mapped_variable, DataConflictAstAttributes::WRITE);
        }
      }
    }
  }

  for(std::list<SymbolWithPath>::iterator
      read_iter  = second_seg.read_variables_.begin(),
      read_end   = second_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if(!tools::is_reference_variable(read_iter->symbol_)) {

      temp.symbol_ = read_iter->symbol_;
      temp.instance_id_ = inst_id_second;

      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }

      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }

      if(tools::is_channel_variable(temp.symbol_)) {
        temp.instance_id_
          = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            (temp.symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute != NULL) {
        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_write_.begin();
            iter != data_conflict_attribute->coloring_write_.end();
            iter++) {

          if(tools::is_member_variable(temp.symbol_)) {

            if(tools::is_static_variable(temp.symbol_)) {
              conflict = true;
            }

            if(temp.instance_id_ == *iter) {
              conflict = true;
            } else {

              conflict |= false;
            }
          } else {
            conflict = true;
          }

          if(debugging_) {
            conflicts.insert(temp);
          } else {
            break;
          }
        }
      }

      conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::WRITE, conflicts);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_second,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {

        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        std::list<MappedVariable> variables;

        if(tools::is_channel_variable(read_iter->symbol_)) {
          //mapped_variable->instance_id_
          //  = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);

          int channel_id
            = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);

          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->port_mapping_cache_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  read_iter->symbol_));

          // Checking the connected path for symbol instance
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {

            int symbol_id = get_id(iter->second, (*mapped_variable).symbol_);

            if (symbol_id != -2) {

              mapped_variable->instance_id_ = symbol_id;
              variables.push_back(*mapped_variable);
            }
          }
        } else {
          variables.push_back(*mapped_variable);
        }

        for(std::list<MappedVariable>::iterator
            var_iter  = variables.begin();
            var_iter != variables.end();
            var_iter++) {

          DataConflictAstAttributes *data_conflict_attribute =
            dynamic_cast<DataConflictAstAttributes*>(
                (var_iter->symbol_)->getAttribute("data_conflict"));

          if(data_conflict_attribute != NULL) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_write_.begin();
                iter != data_conflict_attribute->coloring_write_.end();
                iter++) {

              if(tools::is_member_variable(var_iter->symbol_)) {

                if(tools::is_static_variable(var_iter->symbol_)) {
                  conflict = true;
                }

                if(var_iter->instance_id_ == *iter) {
                  conflict = true;
                } else {

                  conflict |= false;
                }
              } else {
                conflict = true;
              }

              if(debugging_) {
                conflicts.insert((*var_iter));
              } else {
                break;
              }
            }
          }

          conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::WRITE, conflicts);
        }
      }
    }
  }

  for(std::list<SymbolWithPath>::iterator
      write_iter  = second_seg.write_variables_.begin(),
      write_end   = second_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {

    if(!tools::is_reference_variable(write_iter->symbol_)) {

      temp.symbol_ = write_iter->symbol_;
      temp.instance_id_ = inst_id_second;

      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }

      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }

      if(tools::is_channel_variable(temp.symbol_)) {
        temp.instance_id_
          = get_channel_id(*write_iter, temp.instance_id_, dynamic_analysis);
      }

      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            (temp.symbol_)->getAttribute("data_conflict"));

      if(data_conflict_attribute != NULL) {

        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_read_.begin();
            iter != data_conflict_attribute->coloring_read_.end();
            iter++) {

          if(tools::is_member_variable(temp.symbol_)) {

            if(tools::is_static_variable(temp.symbol_)) {
              conflict = true;
            }

            if(temp.instance_id_ == *iter) {

              conflict = true;
            } else {

              conflict |= false;
            }
          } else {
            conflict = true;
          }

          if(debugging_) {
            conflicts.insert(temp);
          } else {
            break;
          }
        }

        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_write_.begin();
            iter != data_conflict_attribute->coloring_write_.end();
            iter++) {

          if(tools::is_member_variable(temp.symbol_)) {

            if(tools::is_static_variable(temp.symbol_)) {
              conflict = true;
            }

            if(temp.instance_id_ == *iter) {

              conflict = true;
            } else {

              conflict |= false;
            }
          } else {
            conflict = true;
          }

          if(debugging_) {
            conflicts.insert(temp);
          } else {
            break;
          }
        }
      }

      conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::WRITE, conflicts);
      conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::READ, conflicts);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_second,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {

        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        std::list<MappedVariable> variables;

        if(tools::is_channel_variable(write_iter->symbol_)) {

          int channel_id
            = get_channel_id(*write_iter, inst_id_second, dynamic_analysis);

          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->port_mapping_cache_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  write_iter->symbol_));

          // Checking the connected path for symbol instance
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {

            int symbol_id = get_id(iter->second, (*mapped_variable).symbol_);

            if (symbol_id != -2) {

              mapped_variable->instance_id_ = symbol_id;
              variables.push_back(*mapped_variable);
            }
          }

        } else {
          variables.push_back(*mapped_variable);
        }

        for(std::list<MappedVariable>::iterator
            var_iter  = variables.begin();
            var_iter != variables.end();
            var_iter++) {

          DataConflictAstAttributes *data_conflict_attribute =
            dynamic_cast<DataConflictAstAttributes*>(
                (var_iter->symbol_)->getAttribute("data_conflict"));

          for(std::set<int>::iterator
              iter  = data_conflict_attribute->coloring_read_.begin();
              iter != data_conflict_attribute->coloring_read_.end();
              iter++) {

            if(tools::is_member_variable(var_iter->symbol_)) {

              if(tools::is_static_variable(var_iter->symbol_)) {
                conflict = true;
              }

              if(var_iter->instance_id_ == *iter) {

                conflict = true;
              } else {

                conflict |= false;
              }
            } else {
              conflict = true;
            }

            if(debugging_) {
              conflicts.insert((*var_iter));
            } else {
              break;
            }
          }

          for(std::set<int>::iterator
              iter  = data_conflict_attribute->coloring_write_.begin();
              iter != data_conflict_attribute->coloring_write_.end();
              iter++) {

            if(tools::is_member_variable(var_iter->symbol_)) {

              if(tools::is_static_variable(var_iter->symbol_)) {
                conflict = true;
              }

              if(var_iter->instance_id_ == *iter) {

                conflict = true;
              } else {

                conflict |= false;
              }
            } else {
              conflict = true;
            }

            if(debugging_) {
              conflicts.insert((*var_iter));
            } else {
              break;
            }
          }

          conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::WRITE, conflicts);
          conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::READ, conflicts);

        }
      }
    }
  }

  for(std::list<SymbolWithPath>::iterator
      read_iter  = first_seg.read_variables_.begin(),
      read_end   = first_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if(!tools::is_reference_variable(read_iter->symbol_)) {

      temp.symbol_ = read_iter->symbol_;
      temp.instance_id_ = MappedVariable::InvalidVariable;

      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }

      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }

      color_symbol(temp, DataConflictAstAttributes::UNUSED);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_first,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {


        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(tools::is_channel_variable((*mapped_variable).symbol_)) {
          mapped_variable->instance_id_
            = get_channel_id(*read_iter, inst_id_first, dynamic_analysis);
        }

        mapped_variable->instance_id_ = MappedVariable::InvalidVariable;
        color_symbol(*mapped_variable, DataConflictAstAttributes::UNUSED);
      }
    }
  }

  for(std::list<SymbolWithPath>::iterator
      write_iter  = first_seg.write_variables_.begin(),
      write_end   = first_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {

    if(!tools::is_reference_variable(write_iter->symbol_)) {

      temp.symbol_ = write_iter->symbol_;
      temp.instance_id_ = MappedVariable::InvalidVariable;

      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }

      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }

      color_symbol(temp, DataConflictAstAttributes::UNUSED);

    } else {

      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_first,
            dynamic_analysis);

      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {

        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }

        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }

        mapped_variable->instance_id_ = MappedVariable::InvalidVariable;
        color_symbol(*mapped_variable, DataConflictAstAttributes::UNUSED);
      }
    }
  }

  if(debugging_) {

    conflict_info_[row][column].conflict_variables_ = conflicts;
    conflict_info_[column][row].conflict_variables_ = conflicts;
  }

  return conflict;
}

bool
risc::sg::DataConflictTable::is_conflict_free_variable(SgVariableSymbol *symbol)
{
  // caching of results
  static std::unordered_map<SgVariableSymbol*, bool> cached_mapping;
  std::unordered_map<SgVariableSymbol*, bool>::iterator iter
    = cached_mapping.find(symbol);
  if(iter != cached_mapping.end()) {
    return iter->second;
  }

  DataConflictAstAttributes *data_conflict_attribute =
    dynamic_cast<DataConflictAstAttributes*>(
        symbol->getAttribute("data_conflict"));

  bool result = false;
  if(data_conflict_attribute == NULL) {

    // this is the case of a mapped variable
    risc::tools::analyze_conflict_free(symbol);
    data_conflict_attribute
      = dynamic_cast<DataConflictAstAttributes*>(
          symbol->getAttribute("data_conflict"));
  }

  result = data_conflict_attribute->is_conflict_free_;

  // caching of results
  cached_mapping[symbol] = result;
  return result;
}

void
risc::sg::DataConflictTable::color_symbol(
      const MappedVariable &variable,
      DataConflictAstAttributes::Status color)
{
  DataConflictAstAttributes *data_conflict_attribute =
    dynamic_cast<DataConflictAstAttributes*>((variable.symbol_)
        ->getAttribute("data_conflict"));

  if(data_conflict_attribute == NULL) {

    data_conflict_attribute = new DataConflictAstAttributes();
    variable.symbol_->setAttribute("data_conflict", data_conflict_attribute);
  }
  if(color == DataConflictAstAttributes::UNUSED) {
    data_conflict_attribute->coloring_read_.clear();
    data_conflict_attribute->coloring_write_.clear();
  } else {
    if(color == DataConflictAstAttributes::READ) {
      data_conflict_attribute->coloring_read_.insert(variable.instance_id_);
    } else {
      data_conflict_attribute->coloring_write_.insert(variable.instance_id_);
    }
  }

  // In the case that we are reading a non pod data type we have to mark
  // all class/struct members as read
  color_members_of_class(variable, color);
}


int get_id(risc::InstanceTree path, SgVariableSymbol *symbol)
{

  // We reduce the path element by element and check if the symbol is
  // in the related module
  while (!path.path_.empty()) {

    // Getting the current variable and the related class definition
    SgVariableDefinition *vardef = path.get_object()->get_ast_node();
    SgClassType *type = isSgClassType(vardef->get_type());

    SgClassDeclaration *classdecl
      = isSgClassDeclaration(type->get_declaration());
    classdecl = isSgClassDeclaration(classdecl->get_definingDeclaration());
    SgClassDefinition *classdef = classdecl->get_definition();


    // check if the symbol is defined in the scope
    if(classdef->symbol_exists(symbol)) {

      return risc::get_design()->path_instance_mapper_->get_id(path);
    }

    // if not found we reduce the instance path by one
    path.path_.pop_back();
  }

  // last check is the global scope
  if(SageInterface::getGlobalScope(symbol)->symbol_exists(symbol)) {
    return -1;
  }

  // The magic value -2 says that the variable is not found
  return -2;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
