#ifndef SEGMENT_GRAPH_H_INCLUDED_
#define SEGMENT_GRAPH_H_INCLUDED_

#include<iostream>
#include<map>
#include<set>
#include<utility>

#include "boost/graph/adjacency_list.hpp"
#include "boost/graph/graphviz.hpp"
#include "boost/algorithm/string.hpp"
#include "../internal_representation/module.h"

#include "rose.h"

#include "edge.h"
#include "function_annotation.h"
#include "port_call_path.h"
#include "segment.h"


#define DBG_SHOW(name,id) std::cout<<"\n\ndebug id: "<< id<<"  "<<#name<<" :  \n"<<name->unparseToString()<<std::endl;
namespace risc {

class Function;

namespace sg {

class SegmentGraph;

typedef boost::adjacency_list<boost::listS,
        boost::listS, boost::bidirectionalS, Segment, Edge,
        boost::property<boost::vertex_index_t, int> > Graph;
typedef boost::graph_traits<Graph>::vertex_descriptor VertexDescriptor;
typedef boost::graph_traits<Graph>::edge_descriptor EdgeDescriptor;

typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;

class FunctionAnnotationAttributes;
class CachedFunctionAstAttributes;

/**
 *  \class SegmentGraph
 *  \brief This class represents a segment graph for a process
 *
 */
class SegmentGraph {

public:

  typedef std::set<SgFunctionDefinition*> FunctionCallBoundarySet;
  typedef std::set<VariantT> CCxxKeywordsBoundarySet;

  SegmentGraph();  
  SegmentGraph(FunctionCallBoundarySet function_boundaries);
  SegmentGraph(CCxxKeywordsBoundarySet keyword_boundaries);
  SegmentGraph(FunctionCallBoundarySet function_boundaries,
      CCxxKeywordsBoundarySet keyword_boundaries);
  typedef std::set<VertexDescriptor> SegmentSet;
  
  void get_wait_and_notifying_events();
  void find_module_segments(std::vector<Module*>& module_definitions_);
  void print_psg_file(std::string file_name);
  void read_psg_from_file(std::string);
  void write_psg_attributes(std::string);
  void replace_function_call_vertex_with_its_psg();
  void remove_dangling_segments();
  void integrate_psg_into_sg(VertexDescriptor, SgSymbol *);
  void count_segments_that_belongs_to_certain_psg();
  
  
  std::vector<std::pair<VertexDescriptor, SgSymbol *> > function_call_segments;  
  std::map< SgSymbol * 	, std::vector<VertexDescriptor> > map_of_function_name_and_starting_segments;
  std::map< SgSymbol * 	, std::vector<VertexDescriptor> > map_of_function_name_and_ending_segments;
  std::vector<SgExprStatement*> vector_of_statements_which_contain_non_defining_func_calls;  
  std::map<std::string, SegmentSet> map_of_function_name_and_leaving_segments_id_to_be_stored_into_a_psg_file;
  std::map<std::string, VertexDescriptor> map_of_function_name_and_starting_segments_id_to_be_stored_into_a_psg_file;
  std::set<VertexDescriptor> first_segments;
  std::map<VertexDescriptor, int> is_psg_first_segment_touched; 
  std::map<VertexDescriptor, std::vector<VertexDescriptor> > psg_segments; 
  std::map<int,int> count_of_segments_that_belongs_to_certain_psg;
  
  /**
   * \fn       clean_graph
   * \brief    This function deletes all duplicated nodes in each vertex
   *           of the graph
   */
  void clean_graph();


  /**
   * \fn       print_graph
   * \brief    This function prints the graph in a dot file.
   *           The name of the file is user defined by argument
   * \param    filename   The name of the file
   */
  void print_graph(std::string filename);


  /**
   * \fn       print_graph_read_write_access
   * \brief    This function prints all variables which have read or written
   *           in this segment.
   * \param    filename   The name of the file
   */
  void print_graph_read_write_access(std::string filename);


  /**
   * \fn       add_function
   * \brief    This function analyzes the given function and creates a
   *           segment graph for it. The graph will be automatically inserted
   *           in the existing graph. The flag duplicate_segments will be
   *           used to generate an instance segement graph.
   *           is_simulation_process: if set to true, meaning the function is a 
   *           simulation process, and it needs a segment id
   */
  void add_function(Function *function, bool duplicate_segments = false, bool is_simulation_process = true, std::string module_name = "");

  

  /**
   * \fn       read_write_analysis_of_segments
   * \brief    The function determines which variables are read and written in
   *           the individual segments
   */
  void read_write_analysis_of_segments();

  /**
   * \fn       set_all_segments_to_untouched
   * \brief    This function sets the status of the flag
   *           helper_for_graph_algorithms for all nodes to 'Untouched'
   */
  void set_all_segments_to_untouched();

private:

  bool static_analysis_;

  /**
   * \fn        handle_recursive_calls
   * \brief     This function integrates the recursive function calls
   */
  void handle_recursive_calls();


  /**
   * \fn        separate_variable_declaration_and_initializer_on_demand
   * \brief     This function checks if a direct or indirect boundary call is
   *            in the initializer expression of the variable declaration.
   *            If yes, the declaration and initialization will be separated.
   *            Example:
   *            Input:
   *                int x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int x;
   *                x = y + foo();
   */
  void separate_variable_declaration_and_initializer_on_demand(
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp);


  /**
   * \fn        decompose_expression_with_boundary_calls
   * \brief     This function decomposes an expression into multiple
   *            statements. Functions which call directly or indirectly a
   *            boundary will be separated and have temporally variables.
   *            Example:
   *            Input:
   *                x = y + foo(); // foo() has a indirect boundary
   *            Output:
   *                int temp_foo = foo();
   *                int x = y + temp_foo();
   */
  void decompose_expression_with_boundary_calls(SgExpression *expr,
           SgBasicBlock *&bb, bool duplicate_segments, PortCallPath pcp);


  /**
   * \fn       create_temp_variable_for_expression
   * \brief    This function creates a variable declaration of the type of the
   *           given expression. The variable will be inserted before the
   *           associated expression statement of the given expression.
   */
  void create_temp_variable_for_expression(SgExpression *expr,
           SgBasicBlock *&bb);


  /**
   * \fn       generate_unique_name
   * \brief    Generates an unique name for a varible based on the given node.
   *
   */
  std::string generate_unique_name(SgNode *node);


  /**
   * \fn         is_boundary_stmt
   * \brief      This function checks if a given statement is a potential
   *             boundary statement
   * \param      current_stmt The statement which will be checked
   */
  bool is_boundary_stmt(SgStatement const * const current_stmt);


  /**
   * \fn       has_function_call_with_boundary
   * \brief    This function checks if in the given expression a function
   *           call occurs which calls a functon with a boundary statements.
   */
  SgFunctionCallExp* has_function_call_with_boundary(SgExpression *expr,
                         bool duplicate_segments, PortCallPath pcp);


  /**
   * \fn       should_decompose_function
   * \brief    This function determines if a expression should be decomposed.
   *           This is the case if expression has the following structure:
   *           var = somepointer->func();
   *           var = somevariable.func();
   *           var = func();
   *           and var is a global, member, or static variable.
   */
  bool should_decompose_function(SgExpression *expr, bool duplicate_segments,
      PortCallPath pcp);


  /**
   * \fn       contains_call_expression
   * \brief    In this function we look for expressions which contain a function
   *           call in the following style:
   *           foo();
   *           this->foo();
   *           var = foo();
   *           var = this->foo();
   *           The return value will be the function declaration of the called
   *           function
   */
  SgFunctionDeclaration* contains_function_call_expression(SgExpression *expr);


  /**
   * \fn       build_graph
   * \brief    This function creates the segment graph
   *
   * \param    current_stmt        The first statement of the sequence which
   *                               will be analyzed
   * \param    current_segments    The beginning segments
   */
  SegmentSet build_graph(
      SgStatement *current_stmt,
      SegmentSet current_segments,
      SegmentSet &break_segments,
      SegmentSet &continue_segments,
      bool duplicate_segments,
      PortCallPath pcp);


  /**
   * \fn       add_expression_to_segment
   * \brief    This function adds to each segment the given expression
   *
   * \param    segments   These segments which will contain the given
   *                      expression
   * \param    expr       The expressions which will added to all segments
   */
  void add_expression_to_segment(SegmentSet segments, SgNode *expr,
      PortCallPath pcp);


  /**
   * \fn       build_segment_graph_for_function
   * \brief    This function generates a segment graph for a function
   */
  CachedFunctionAstAttributes*
  build_segment_graph_for_function(
      SgFunctionDefinition *func_def,
      bool duplicate_segments,
      PortCallPath pcp);


  /**
   * \fn       insert_loop_edges
   * \brief    This is a helper function for loop handling
   *
   */
  void insert_loop_edges(VertexDescriptor &loop_vertex,
           SegmentSet &current_segments,
           SegmentSet &continue_segments,
           SegmentSet &leaf_segments_of_loop);

  /**
   * \fn    analyze_function_call_annotation
   * \brief This function takes a function call expression and checks if there
   *        is any function annotation at the declaration. If the function
   *        declaration has no annotation, there will be a check if a annotation
   *        must be created
   *
   *        FA: Function annoation
   *        CF: Cached function
   *
   *                           comment        no comment
   *        definition      FA (inlining)    CF (linking)
   *        no defiition    FA (inlining)    FA (nothing,
   *                                         default: no-wait conflict-free)
   */
  FunctionAnnotationAttributes* get_function_call_annotation(
      SgFunctionCallExp *func_call_exp);

  /**
   * \fn    create_cached_function_attribute_for_annotated_function
   * \brief This function creates a cached function annotation information
   *        for based on the cached function information
   */
  CachedFunctionAstAttributes*
  create_cached_function_attribute_for_annotated_function(
      SgNode* boundary_node,
      WAIT_CONSTRUCT wait_type,
      SegmentSet incoming_segments,
      bool is_conflict_free,
      PortCallPath pcp);

  /**
   * \fn    add_mapped_symbol_to_reference
   * \brief Let's take a look at this example:
   *        void foo(int &ref, int value, in &ref2);
   *        foo(var, some_var, other_var);
   *        This function adds the symbol of "var" to the attribute "references"
   *        of the symbol "ref".
   *        The same happens for "ref2" and "other_var".
   */
  void add_mapped_symbol_to_reference(SgFunctionCallExp *func_call,
           SgFunctionDefinition *func_def);

  /**
   * \fn    forward_reference_function_parameters
   * \brief This function forwards the arguments which are passed by reference
   *        to the function parameters of the called function. Recursively,
   *        this function traverses through the body of the function definition
   *        and follow other function calls.
   */
  void forward_reference_function_parameters(SgFunctionCallExp *func_call_exp);

  /**
   * \fn    check_and_mark_not_conflict_free_func_call
   * \brief This function checks if there is a not-conflict-free function call
   *        in the expression. If so, all segments marked as in conflict.
   */
  void mark_not_conflict_free_func_calls_in_segment(
           SgNode *stmt,
           SegmentSet segments);


  /**
   * \fn    analyze_prepended_pragmas_for_annotation
   * \brief This function checks if the function has any pragma.
   *        The results are in the return value
   */
  FunctionAnnotation*
  analyze_prepended_pragmas_for_annotation(SgFunctionDeclaration *func_decl);

  SegmentSet duplicate_segments(SegmentSet segments);
  SegmentSet duplicate_empty_segments(SegmentSet segments);

  FunctionCallBoundarySet function_boundaries_;
  CCxxKeywordsBoundarySet keyword_boundaries_;

  std::set<VertexDescriptor> recursive_function_calls_;

  /**
   * \brief This counter is used for generating variables names with unique
   *        names
   */
  static int counter;

  /**
   * \fn    follow_function_calls
   * \brief true, if the function be followed.
   */
  bool follow_function_call(SgFunctionDeclaration *func_decl);

public:

  /**
   * \fn    has_transition
   * \brief true, if there is a transition from 'from' to 'to'
   */
  bool has_transition(int seg_id_from, int seg_id_to);

  /**
   * \fn    id_to_vertex_descriptor
   * \brief Returns a vertex descriptor of the segment with the id 'seg_id'
   */
  VertexDescriptor id_to_vertex_descriptor(int seg_id);

  /**
   * \brief Internal data structure of the segment graph
   */
  Graph graph_;

  /**
   * \brief This set stores segments from precached functions.
   *        These segments will be deleted later.
   */
  std::set<VertexDescriptor> precached_function_segments_;

  /**
   * \brief This set stores all function calls which belong to a port.
   *        These segments will be deleted later
   */
  std::set<SgFunctionCallExp*> port_calls_;


  /**
   * \brief Is true if the segment graph is successfully created
   */
  bool is_succesfully_created_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_GRAPH_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
